;;;; uri-test.scm

(use testbase testbase-output-human)
(use uri)

#| TestBase
  (test
    (purpose "Checkout uri build.")
    (feature uri)
    (reader-extension uri))

  (compile
    (extend (use uri)))
|#

;;; Tests

;;

(define-test uri-non-strict-test "URI Non-strict Parsing"
  (initial
    (define turi #f)
  )

  (side-effect (uri-rfc3986-strict #f))

  (expect-set! turi (uri "foo.org"))
  (expect-true (uri? turi))
  (expect-equal "foo.org" (uri-host turi))
  (expect-false (uri-path turi))
  (expect-false (uri-username turi))
  (expect-false (uri-password turi))
  (expect-equal "//foo.org" (uri->string turi))

  (expect-equal "Generic components"
    '(#f ((#f "<path>" #f) #f (("<query>" . #t))) "<fragment>")
    (uri->list "<path>?<query>#<fragment>"))

  (expect-equal
    '(urn ((#f "example:animal:ferret" #f) #f #f) #f)
    (uri->list "urn:example:animal:ferret:nose"))

  (expect-equal
    '(urn ((#f "oasis:names:specification:docbook:dtd:xml" #f) #f #f) #f)
    (uri->list "urn:oasis:names:specification:docbook:dtd:xml:4.1.2"))

  (expect-equal
    '(mailto (("John.Doe" "example.com" #f) #f #f) #f)
    (uri->list "mailto:John.Doe@example.com"))

  (expect-equal
    '(news ((#f "comp.infosystems.www.servers.unix" #f) #f #f) #f)
    (uri->list "news:comp.infosystems.www.servers.unix"))

  (expect-equal
    '(tel ((#f "+1-816-555-1212" #f) #f #f) #f)
    (uri->list "tel:+1-816-555-1212"))

  (side-effect (uri-rfc3986-strict #t))
)

;;

(define-test uri--strict-test "URI Strict Parsing"
  (initial
    (define turi #f)
  )

  (expect-set! turi (uri "foo.org"))
  (expect-true (uri? turi))
  (expect-equal '("foo.org") (uri-path turi))
  (expect-false (uri-host turi))
  (expect-false (uri-username turi))
  (expect-false (uri-password turi))
  (expect-equal "foo.org" (uri->string turi))

  (expect-set! turi
    (make-uri #:scheme 'http #:authority '(#f "www.ics.uci.edu" 80) #:path '("" "pub" "ietf" "uri" "") #:fragment "Related"))
  (expect-true (uri? turi))
  (expect-equal "http://www.ics.uci.edu:80/pub/ietf/uri/#Related" (uri->string turi))

  (expect-equal
    '(http ((#f "www.ics.uci.edu" 80) ("" "pub" "ietf" "uri" "") #f) "Related")
    (uri->list "http://www.ics.uci.edu:80/pub/ietf/uri/#Related"))

  (expect-equal (warn "Should the path really be (\"\" \"\") & not (\"\")?")
    '(http ((#f "www.call-with-current-continuation.org" #f) ("" "") #f) #f)
    (uri->list "http://www.call-with-current-continuation.org/"))

  (expect-equal "Generic components"
    '(#f (#f ("<path>") (("<query>" . #t))) "<fragment>")
    (uri->list "<path>?<query>#<fragment>"))

  (expect-equal "Generic components"
    '(#f ((#f "<authority>" #f) ("" "<path>") (("<query>" . #t))) "<fragment>")
    (uri->list "<authority>/<path>?<query>#<fragment>"))

  (expect-equal (warn "This is an invalid URI but it still parses!")
    '(ftp (("cnn.example.com&story=breaking_news" "10.0.0.1" #f) ("" "top_story.htm") #f) #f)
    (uri->list "ftp://cnn.example.com&story=breaking_news@10.0.0.1/top_story.htm"))

  (expect-equal
    '(ftp ((#f "ftp.is.co.za" #f) ("" "rfc" "rfc1808.txt") #f) #f)
    (uri->list "ftp://ftp.is.co.za/rfc/rfc1808.txt"))

  (expect-equal "Parse of an ldap uri scheme-specific-part"
    '(ldap ((#f "[2001:db8::7]" #f) ("" "c=GB") (("objectClass?one" . #t))) #f)
    (uri->list "ldap://[2001:db8::7]/c=GB?objectClass?one"))

  (expect-equal
    '(urn (#f ("example:animal:ferret:nose") #f) #f)
    (uri->list "urn:example:animal:ferret:nose"))

  (expect-equal
    '(urn (#f ("oasis:names:specification:docbook:dtd:xml:4.1.2") #f) #f)
    (uri->list "urn:oasis:names:specification:docbook:dtd:xml:4.1.2"))

  (expect-equal
    '(mailto (#f ("John.Doe@example.com") #f) #f)
    (uri->list "mailto:John.Doe@example.com"))

  (expect-equal
    '(news (#f ("comp.infosystems.www.servers.unix") #f) #f)
    (uri->list "news:comp.infosystems.www.servers.unix"))

  (expect-equal
    '(tel (#f ("+1-816-555-1212") #f) #f)
    (uri->list "tel:+1-816-555-1212"))

  (expect-equal
    '(telnet ((#f "192.0.2.16" 80) ("" "") #f) #f)
    (uri->list "telnet://192.0.2.16:80/"))
)

;;

(define-test uri-query-test "URI Query"
  (initial
    (define turi #f)
  )

  (expect-set! turi
    (uri "http://www.ics.uci.edu:80/pub/ietf/uri?foo=23+45&bar=abcedf&bax=x%0Ay"))
  (expect-equal (warn "Note reversal of query items turi!")
    '(("bax" . "x\ny") ("bar" . "abcedf") ("foo" . "23 45"))
    (uri-query turi))
  (expect-equal (warn "Note reversal of query items from turi!")
    "http://www.ics.uci.edu:80/pub/ietf/uri?bax=x%0Ay&bar=abcedf&foo=23+45"
    (uri->string turi))

  (expect-success "Query w/ ?"
    (uri "http://www.ics.uci.edu:80/pub/ietf/uri?foo?bar"))

  (expect-equal "Unencoded query as parameter"
    '(("bax" . "x\ny") ("bar" . "abcedf") ("foo" . "23 45"))
    (uri-query (make-uri #:query "foo=23 45&bar=abcedf&bax=x\ny")))

  (expect-equal "Encoded query"
    "bax=x%0Ay&bar=abcedf&foo=23+45"
    (uri-encode-query (uri-query (make-uri #:query "foo=23 45&bar=abcedf&bax=x\ny"))))

  (test/case "Query key w/o value - daishi@axlight.com"
    (expect-success  (uri "http://1.1.1.1/abc?def"))
    (expect-equal '(("pbs" . foo) ("abc" . "cbs")) (uri-decode-query "abc=cbs&pbs" 'foo))
    (expect-equal "pbs&abc=cbs" (uri-encode-query '(("pbs" . foo) ("abc" . "cbs")) 'foo))
  )

  (expect-equal "Encoded query w/ special chars"
    "foo=test+%26+co&bar=quux%2Bbla"
    (uri-encode-query '(("foo" . "test & co") ("bar" . "quux+bla"))))

  (expect-equal "Decoded query w/ special chars"
    '(("bar" . "quux+bla") ("foo" . "test & co"))
    (uri-decode-query "foo=test+%26+co&bar=quux%2Bbla"))
)

;;

(define-test uri-utility-test "URI Utility"
  (initial
    (define turi #f)
  )

  (expect-false "abs uri 1" (absolute-uri? ":"))
  (expect-false "abs uri 2" (absolute-uri? "abc:"))
  (expect-false "abs uri 3" (absolute-uri? ":cbs"))
  (expect-true "abs uri 3" (absolute-uri? "abc:cbs"))
)

;;

(define (uri/basis uri bas)
  (uri->string (list->uri (uri->list uri bas))) )

(define-test uri-relative-test "URI Relative"
  (initial
    (define basis (uri->list "http://a/b/c/d;p?q")) )

  (expect-equal "g:h"
                (uri/basis "g:h" basis))
  (expect-equal "http://a/b/c/g"
								(uri/basis "g" basis))
  (expect-equal "http://a/b/c/g"
								(uri/basis "./g" basis))
  (expect-equal "http://a/b/c/g/"
								(uri/basis "g/" basis))
  (expect-equal "http://a/g"
								(uri/basis "/g" basis))
  (expect-equal "http://g"
								(uri/basis "//g" basis))
  (expect-equal "http://a/b/c/d;p?y"
								(uri/basis "?y" basis))
  (expect-equal "http://a/b/c/g?y"
								(uri/basis "g?y" basis))
  (expect-equal "http://a/b/c/d;p?q#s"
								(uri/basis "#s" basis))
  (expect-equal "http://a/b/c/g#s"
								(uri/basis "g#s" basis))
  (expect-equal "http://a/b/c/g?y#s"
								(uri/basis "g?y#s" basis))
  (expect-equal "http://a/b/c/;x"
								(uri/basis ";x" basis))
  (expect-equal "http://a/b/c/g;x"
								(uri/basis "g;x" basis))
  (expect-equal "http://a/b/c/g;x?y#s"
								(uri/basis "g;x?y#s" basis))
  (expect-equal "http://a/b/c/d;p?q"
								(uri/basis "" basis))
  (expect-equal "http://a/b/c/"
								(uri/basis "." basis))
  (expect-equal "http://a/b/c/"
								(uri/basis "./" basis))
  (expect-equal "http://a/b/"
								(uri/basis ".." basis))
  (expect-equal "http://a/b/"
								(uri/basis "../" basis))
  (expect-equal "http://a/b/g"
								(uri/basis "../g" basis))
  (expect-equal "http://a/"
								(uri/basis "../.." basis))
  (expect-equal "http://a/"
								(uri/basis "../../" basis))
  (expect-equal "http://a/g"
								(uri/basis "../../g" basis))

  (expect-equal "http://a/g"
								(uri/basis "../../../g" basis))
  (expect-equal "http://a/g"
								(uri/basis "../../../../g" basis))

  (expect-equal "http://a/g"
								(uri/basis "/./g" basis))
  (expect-equal "http://a/g"
								(uri/basis "/../g" basis))
  (expect-equal "http://a/b/c/g."
								(uri/basis "g." basis))
  (expect-equal "http://a/b/c/.g"
								(uri/basis ".g" basis))
  (expect-equal "http://a/b/c/g.."
								(uri/basis "g.." basis))
  (expect-equal "http://a/b/c/..g"
								(uri/basis "..g" basis))

  (expect-equal "http://a/b/g"
								(uri/basis "./../g" basis))
  (expect-equal "http://a/b/c/g/"
								(uri/basis "./g/." basis))
  (expect-equal "http://a/b/c/g/h"
								(uri/basis "g/./h" basis))
  (expect-equal "http://a/b/c/h"
								(uri/basis "g/../h" basis))
  (expect-equal "http://a/b/c/g;x=1/y"
								(uri/basis "g;x=1/./y" basis))
  (expect-equal "http://a/b/c/y"
								(uri/basis "g;x=1/../y" basis))

  (expect-equal "http://a/b/c/g?y/./x"
								(uri/basis "g?y/./x" basis))
  (expect-equal "http://a/b/c/g?y/../x"
								(uri/basis "g?y/../x" basis))
  (expect-equal "http://a/b/c/g#s/./x"
								(uri/basis "g#s/./x" basis))
  (expect-equal "http://a/b/c/g#s/../x"
								(uri/basis "g#s/../x" basis))

  (expect-equal "http:g"
								(uri/basis "http:g" basis))
)

;;; Main

(test::for-each (cut test::styler-set! <> test::output-style-human))
(run-test "URI Tests")

(test::forget!)
