;-*-scheme-*-
;;;; uri.scm
;;;; Kon Lovett, Sep '06

; Portions originally from:
;;; "uri.scm" Construct and decode Uniform Resource Identifiers.
; Copyright 1997, 1998, 2000, 2001 Aubrey Jaffer
;
;Permission to copy this software, to modify it, to redistribute it,
;to distribute modified versions, and to use it for any purpose is
;granted, subject to the following restrictions and understandings.
;
;1.  Any copy made of this software must include this copyright notice
;in full.
;
;2.  I have made no warranty or representation that the operation of
;this software will be error-free, and I am under no obligation to
;provide any services, by way of maintenance, update, or otherwise.
;
;3.  In conjunction with products arising from the use of this
;material, there shall be no use of my name in any advertising,
;promotional, or sales literature without prior written consent in
;each case.

(use srfi-1 #;srfi-12 srfi-13 files regex posix)
(use coerce)        ; duh
(use miscmacros)    ; For define-parameter
(use srfi-18 synch) ; For thread-safe scheme-specific table
(use lookup-table)  ; For efficient scheme-specific table
;(use misc-extn-condition) ; For condition API

(eval-when (compile)
  (declare
    (usual-integrations)
    (fixnum)
    (inline)
    (lambda-lift)
    (no-procedure-checks)
    (bound-to-procedure
      *default-scheme-specific-class*)
    (export
      ;;
      string-ci->symbol
      ;;
      #;exn-uri-query-parse-condition?
      ;;
      uri-missing-query-value
      uri-rfc3986-strict
      ;;
      uri make-uri *make-uri
      uri?
      uri-scheme
      uri-specific
      uri-fragment
      uri-field
      uri-authority
      uri-userinfo
      uri-username
      uri-password
      uri-host
      uri-port
      uri-path
      uri-query
      uri->string
      uri->list list->uri
      uri-encode uri-decode
      uri-decode-split-path uri-encode-split-path
      uri-decode-query uri-encode-query
      uri-decode-query-item uri-encode-query-item
      uri-decode-path
      pathname->uri
      absolute-uri?
      uri-split-path
      uri-split-path-null?
      uri-path-null?
      uri-scheme-symbol
      uri-new-scheme-specific
      ;; internal
      uri$scheme-set!
      uri$specific-set!
      uri$fragment-set!
      uri$scheme-class
      uri$scheme-class-set!
      uri$scheme-data
      uri$scheme-data-set!
      ;; DEPRECATED
      make-encoded-uri
      uri-scheme-specific
      uri->tree
      uri-make-path
      path->uri
      uri-split-fields
      uri-path->keys
      absolute-path?
      null-directory?
      glob-pattern?) ) )

;;;

#;
(define (char-hex-numeric? chr)
  (string-index "ABCEDEFabcdef0123456789" chr) )

;;;
;;; URI Record
;;;

(define-record-type uri
  (uri$make scheme specific fragment scheme-class scheme-data)
  uri?
  ; Basic info
  (scheme         uri-scheme        uri$scheme-set!)
  (specific       uri-specific      uri$specific-set!)
  (fragment       uri-fragment      uri$fragment-set!)
  ; Per scheme implementation
  (scheme-class   uri$scheme-class  uri$scheme-class-set!)
  (scheme-data    uri$scheme-data   uri$scheme-data-set!) )

;;;
;;; Private
;;;

;; Constants

(define-constant RFC3986-GEN-DELIMS ":/?#[]@")
(define-constant RFC3986-SUB-DELIMS "!$&'()*+,;=")

(define-constant SCHEME-PATT       "([a-zA-Z][a-zA-Z0-9+.-]*)")
(define-constant SCHEME-PARSE-PATT "([a-zA-Z][a-zA-Z0-9+.-]*):")

(define-constant OPAQUE-PATT "([^\\s]+)")

(define-constant ROOT-SPLIT-PATH '(""))

(define-constant EMPTY-USERINFO-LIST '(#f #f))

(define-constant EMPTY-SCHEME-SPECIFIC '(#f #f #f))

(define-constant EMPTY-URI-ARGS '(#f (#f #f #f) #f))

;; Generic Routines

(define (string-ci->symbol str)
  (string->symbol (string-downcase str)) )

(define (check-procedure obj loc)
  (unless (procedure? obj)
    (error loc "invalid procedure" obj) ) )

(define (check-list obj loc)
  (unless (list? obj)
    (error loc "invalid list" obj) ) )

(define (check-symbol obj loc)
  (unless (symbol? obj)
    (error loc "invalid symbol" obj) ) )

;; URI Conditions

#|
(define (exn-uri-query-parse-condition? obj)
  (and ((condition-predicate 'exn) obj)
       ((condition-predicate 'uri) obj)
       ((condition-predicate 'query) obj)
       ((condition-predicate 'parse) obj) ) )

(define (make-exn-uri-query-parse-condition location message arguments string)
  (make-composite-condition
    (make-property-condition 'exn 'location location 'message message 'arguments arguments)
    (make-property-condition 'uri)
    (make-property-condition 'query)
    (make-property-condition 'parse 'string string)) )
|#

;; Scheme Specific Support

(define (make-scheme-specific name
            #!key
            fielder
            parser
            (maker (lambda args (first args)))
            encoder
            decoder
            data
            (associations '()))
  (check-procedure fielder 'make-scheme-specific)
  (check-procedure parser 'make-scheme-specific)
  (check-procedure maker 'make-scheme-specific)
  (check-procedure encoder 'make-scheme-specific)
  (check-procedure decoder 'make-scheme-specific)
  (check-list associations 'make-scheme-specific)
  (unless (null? associations)
    (for-each (lambda (cell) (check-procedure (cdr cell) 'make-scheme-specific)) associations) )
  (letrec ([specific-dispatcher
            (lambda (op . args)
              (switch op
                ['field
                  (let ([uri (first args)]
                        [key (second args)])
                    (switch key
                      ['scheme (uri-scheme uri)]
                      ['specific (uri-specific uri)]
                      ['fragment (uri-fragment uri)]
                      [else
                        (apply fielder args)]))]
                ['make
                  (apply maker
                    (uri$make
                      (first args) (second args) (third args)
                      specific-dispatcher (specific-dispatcher 'data))
                    args)]
                ['parse
                  (apply parser args)]
                ['decode
                  (apply decoder args)]
                ['encode
                  (apply encoder args)]
                ['name
                  name]
                ['data
                  (if (or (null? args) (not (car args)))
                      data
                      (uri$scheme-data (car args)))]
                [else
                  (let ([proc (alist-ref op associations eq?)])
                    (if proc
                        (apply proc args)
                        (error name "unknown scheme specific operation" name op) ) )]) )] )
    specific-dispatcher ) )

(define *default-scheme-specific-class* #f)

(define *scheme-specific-classes*
  (make-object/synch (make-dict eq? 4) 'scheme-specific-classes))

(define (scheme-specific-class scheme . specific-arg)
  (%let/synch ([sc *scheme-specific-classes*])
    (let ([curr (dict-ref sc scheme *default-scheme-specific-class*)])
      (unless (null? specific-arg)
        (dict-set! sc scheme (first specific-arg)) )
      curr ) ) )

;;

(define-parameter uri-missing-query-value
  #t
  (lambda (x) x) )

;;

(define-parameter uri-rfc3986-strict
  #t
  (lambda (x)
    (if (boolean? x)
        x
        (begin
          (warning 'uri-rfc3986-strict "invalid boolean:" x)
          (uri-rfc3986-strict) ) ) ) )

;; RFC 3986 unreserved?

(define (rfc3986-unreserved-char? chr)
  (or (char-alphabetic? chr)
      (char-numeric? chr)
      (string-index "-_.~" chr)) )

;; RFC 3986 pchar?

(define (rfc3986-path-unreserved-char? chr)
  (or (string-index RFC3986-SUB-DELIMS chr)
      (string-index ":@" chr)
      #;(rfc3986-unreserved-char? chr) ) )

;;

(define (rfc3986-query-unreserved-char? chr)
  (or #;(rfc3986-unreserved-char? chr)
      (string-index "$!*'()," chr) ) )

;; Split path root routines

(define (root-split-path? path-list)
  (equal? ROOT-SPLIT-PATH path-list) )

(define (rooted-split-path? path-list)
  (and (pair? path-list)
       (string=? "" (car path-list))) )

(define (absolute-split-path? path-list)
  (and (rooted-split-path? path-list)
       (not (root-split-path? path-list))) )

;; Returns list from string userinfo.

(define (split-userinfo userinfo)
  (let ([cdx (string-index-right userinfo #\:)])
    (if cdx
        (list
          (substring userinfo 0 cdx)
          (substring userinfo (add1 cdx)))
        (list
          userinfo
          #f) ) ) )

;; Splits an authority string form into a list form,
;; (userinfo host port) where any can be #f.

(define (split-authority authority loc)
  (let* ([cadx (string-index-right authority #\@)]
         [userinfo (and cadx (substring authority 0 cadx))]
         [hostport (if cadx (substring/shared authority (add1 cadx)) authority)]
         [ipv6-end (string-index-right hostport #\])]
         [cdx
           (if ipv6-end
               (if (string-index hostport #\[)
                   (string-index hostport #\: ipv6-end)
                   (error loc "invalid IPv6 address" authority))
               (string-index-right hostport #\:))]
         [host (if cdx (substring hostport 0 cdx) hostport)]
         [port (and cdx (substring hostport (add1 cdx)))])
    (list userinfo host (and port (string->number port))) ) )

;; Splits an string query into an alist.

(define (split-query query default)
  (let ([end (string-length query)])
    (let loop ([cur 0] [alst '()])
      (if (>= cur end)
          alst
          (let* ([ptr (or (string-index query #\& cur end) end)]
                 [edx (string-index query #\= cur ptr)])
            (loop (add1 ptr)
                  (if edx
                      (alist-cons (substring/shared query cur edx)
                                  (substring/shared query (add1 edx) ptr)
                                  alst)
                      (alist-cons (substring/shared query cur ptr)
                                  default
                                  alst)) ) ) ) ) ) )

;; Is the object a valid query pair?

(define (query-pair? obj)
  (and (pair? obj)
       ; Should be string but can be anything
       #;(string? (car obj))
       ; Can be anything
       #;(string? (cdr obj)) ) )

;;;
;;; Public
;;;

;;@body Returns a copy of the string @1 in which all @dfn{unsafe} octets
;;(as defined in RFC 2396) have been @samp{%} @dfn{escaped}.
;;@code{uri-decode} decodes strings encoded by @0.
;;
;; RFC 3986 - % UCASE-HEX-DIGIT UCASE-HEX-DIGIT

(define (uri-cons-encode-char chr lst)
  (let* ([int (char->integer chr)]
         [str (number->string int 16)])
    (cons #\%
          (let ([uc0 (char-upcase (string-ref str 0))])
            (if (< int #x10)
                (cons* #\0 uc0 lst)
                (cons* uc0 (char-upcase (string-ref str 1)) lst) ) ) ) ) )

#; ;UNUSED
(define (uri-cons-encoded-linebreak lst)
  (cons* #\% #\0 #\D #\% #\0 #\A lst) )

#; ;UNUSED
(define (exploded-linebreak? lst)
  (and (<= 2 (length lst))
       (let ([1st (car lst)]
             [2nd (cadr lst)])
         (or (and (char=? #\return 1st)
                  (char=? #\newline 2nd))
             (and (char=? #\newline 1st)
                  (char=? #\return 2nd)))) ) )

(define (uri-encode ustr . args)
  (let ([alloweds
          (map! (lambda (obj)
                  (cond [(procedure? obj)
                           obj]
                        [(string? obj)
                          (lambda (chr) (string-index obj chr))]
                        [(char? obj)
                          (lambda (chr) (char=? obj chr))]
                        [else
                          (error 'uri-encode "invalid allowed test:" obj)]) )
                args) ] )
    (let loop ([ulst (reverse! (string->list ustr))] [elst '()])
      (if (null? ulst)
          (list->string elst)
          (let ([chr (car ulst)]
                [nxt (cdr ulst)])
            (cond [(rfc3986-unreserved-char? chr)
                    (loop nxt (cons chr elst))]
                  [(any (cut <> chr) alloweds)
                    (loop nxt (cons chr elst))]
                  #; ;USELESS
                  [(exploded-linebreak? ulst)
                    (loop (cddr ulst) (uri-cons-encoded-linebreak elst))]
                  #; ;QUESTIONABLE
                  [(char=? #\newline chr)
                    (loop nxt (uri-cons-encoded-linebreak elst))]
                  [else
                    (loop nxt (uri-cons-encode-char chr elst) ) ] ) ) ) ) ) )

;;@body Returns a copy of the string @1 in which each @samp{%} escaped
;;characters in @1 is replaced with the character it encodes.  This
;;routine is useful for showing URI contents on error pages.

(define (uri-decode estr)
  (let ((re (regexp "%[0-9a-fA-F][0-9a-fA-F]")))
    (let loop ([idx 0] [lst '()])
      (match (string-search-positions re estr idx)
        [((start-idx end-idx))
         (loop end-idx
               (cons* (string
                       (integer->char
                        (string->number
                         (substring/shared estr (add1 start-idx) end-idx)
                         16)))
                      (substring/shared estr idx start-idx) lst))]
        [_ (if (null? lst)
               estr
               (string-concatenate-reverse
                (cons (substring/shared estr idx) lst)))] ) )) )

;; Splits a string form of a path into a list.
;; Returns a rooted split path.
;; Only for use when parsing an unencoded string path!

(define (uri-split-path path)
  (let ([path-list (string-split path "/" #t)])
    (if (rooted-split-path? path-list)
        path-list
        (cons "" path-list)) ) )

;; Returns a decoded split path, where any instances
;; of "." & ".." are properly handled.

(define (uri-decode-split-path path-list #!optional base-path-list)
  (let ([path-list (map uri-decode path-list)])
    (cond [(absolute-split-path? path-list)
            path-list]
          [base-path-list
            (let* (
                [cpath0
                  (append (butlast base-path-list) path-list)]
                [cpath1
                  (let loop ([l cpath0] [result '()])
                    (cond [(null? l)
                            (reverse! result)]
                          [(not (equal? "." (car l)))
                            (loop (cdr l) (cons (car l) result))]
                          [(null? (cdr l))
                            (reverse! (cons "" result))]
                          [else
                            (loop (cdr l) result)]))]
                  [cpath2
                    (let loop ([l cpath1] [result '()])
                      (cond [(null? l)
                              (reverse! result)]
                            [(not (equal? ".." (car l)))
                              (loop (cdr l) (cons (car l) result))]
                            [(or (null? result) (string-null? (car result)))
                              (warning 'uri-decode-split-path "null path" cpath1)
                              (append! (reverse! result) l)]
                            [(null? (cdr l))
                              (reverse! (cons "" (cdr result)))]
                            [else
                              (loop (cdr l) (cdr result))]))])
              cpath2)]
          [else
            path-list] ) ) )

;;@body
;;Returns a URI string combining the components of list @1.

(define (uri-encode-split-path path-list)
  (apply string-append
         (uri-encode (car path-list) rfc3986-path-unreserved-char?)
         (map (lambda (elm) (string-append "/" (uri-encode elm rfc3986-path-unreserved-char?)))
              (cdr path-list))) )

;;@body
;;Returns decoded split path from a string path.

(define (uri-decode-path path . args)
  (apply uri-decode-split-path (string-split path "/" #t) args) )

;; Split & uri decode

(define (uri-decode-authority authority)
  (let* ([lst (split-authority authority 'uri-decode-authority)]
         [ui (car lst)])
    (if ui
        (cons (uri-decode ui) (cdr lst))
        lst) ) )

;; Join & uri encode

(define (uri-encode-authority authority-list)
  (apply
    (lambda (userinfo host port)
      (if (or userinfo host port)
          (string-append
            "//"
            (if userinfo (string-append (uri-encode userinfo RFC3986-SUB-DELIMS #\:) "@") "")
            (or host "") ; should exist, but...
            (if port (string-append ":" (number->string port)) ""))
          "") )
      authority-list) )

;; '+' -> ' '

(define (uri-decode-query-item obj)
  (if (string? obj)
      (uri-decode (string-translate obj #\+ #\space))
      obj) )

;; ' ' -> '+'

(define (uri-encode-query-item obj)
  (string-translate (uri-encode (->string obj) rfc3986-query-unreserved-char? #\space)
                    #\space #\+) )

;;@body Converts a @dfn{URI} encoded @1 to a query-alist.

(define (uri-decode-query query #!optional (missing (uri-missing-query-value)))
  (map! (lambda (cell)
          (and-let* ([key (car cell)])
            (set-car! cell (uri-decode-query-item key))
            (set-cdr! cell (uri-decode-query-item (cdr cell))) )
          cell )
        (split-query query missing)) )

;; Returns the string form of a query association list.

(define (uri-encode-query query-alist #!optional (missing (uri-missing-query-value)))
  (string-intersperse
    (map (lambda (cell)
           (let ([key (uri-encode-query-item (car cell))]
                 [binval (cdr cell)])
             (if (eqv? binval missing)
                 key
                 (string-append key "=" (uri-encode-query-item binval)) ) ) )
         query-alist)
    "&") )

;; Returns the leading scheme symbol.

(define uri-scheme-symbol
  (let ([scheme-re (regexp (string-append "^" SCHEME-PATT "$"))]
        [scheme-parse-re (regexp (string-append "^" SCHEME-PARSE-PATT))])
    (lambda (str #!optional parse?)
      (and-let* ([lst (string-search (if parse? scheme-parse-re scheme-re) str)])
        (and (= 2 (length lst))
             (first lst)
             (second lst)
             (string-ci->symbol (second lst))) ) ) ) )

;; Builtin field reference

(define (*uri-field uri key)
  ((uri$scheme-class uri) 'field uri key) )

(define (uri-authority uri)
  (*uri-field uri 'authority) )

(define (uri-path uri)
  (*uri-field uri 'path) )

(define (uri-query uri)
  (*uri-field uri 'query) )

(define (uri-userinfo uri)
  (*uri-field uri 'userinfo) )

(define (uri-username uri)
  (*uri-field uri 'username) )

(define (uri-password uri)
  (*uri-field uri 'password) )

(define (uri-host uri)
  (*uri-field uri 'host) )

(define (uri-port uri)
  (*uri-field uri 'port) )

;; Scheme specific field reference

(define (uri-field uri key . rest)
  (apply (uri$scheme-class uri) 'field uri key rest) )

;; Add a scheme specific handler

(define (uri-new-scheme-specific scheme . rest)
  (check-symbol scheme 'uri-new-scheme-specific)
  (scheme-specific-class scheme (apply make-scheme-specific scheme rest)) )

;; Make a uri object from all required parameters

(define (*make-uri scheme specific fragment . rest)
  (apply (scheme-specific-class scheme) 'make scheme specific fragment rest) )

;; Make a uri object from argument list

(define (make-uri #!rest rest #!key scheme specific fragment)
  (let ([scheme-symbol
          (and scheme
               (or (uri-scheme-symbol
                       (cond [(symbol? scheme) (symbol->string scheme)]
                             [(string? scheme) scheme]
                             [else
                               (error 'make-url "invalid scheme" scheme)]))
                   (error 'make-url "invalid scheme" scheme) ) ) ] )
    (apply *make-uri
           (apply (scheme-specific-class scheme-symbol) 'parse
                  #:scheme scheme-symbol #:specific specific #:fragment fragment
                  rest)) ) )

;; Make a uri object from a string

(define (uri str #!optional default)
  (let ([lst
          ((scheme-specific-class (uri-scheme-symbol str #t)) 'decode str #f)])
    (if lst
        (apply *make-uri lst)
        (or default
            (error 'uri "invalid URI" str) ) ) ) )

;; Make a string from a uri object

(define (uri->string uri)
  ((uri$scheme-class uri) 'encode uri) )

;; Returns the list (scheme specific fragment ...) for the
;; supplied string uri.

(define (uri->list str #!optional basis)
  (let ([basis (cond [(string? basis)
                       (uri->list basis)]
                     [(list? basis)
                       basis]
                     [(not basis)
                       #f]
                     [else
                       (error 'uri->list "invalid basis:" basis)])])
    ((scheme-specific-class (uri-scheme-symbol str #t)) 'decode str basis) ) )

;; Returns a uri object from the unencoded list form

(define (list->uri lst)
  (apply *make-uri lst) )

;;@body Returns a URI-string for @1 on the local host.

(define (pathname->uri path)
  (if (absolute-pathname? path)
      (string-append "file://" path)
      (string-append "file://" (current-directory) "/" path)) )

;;@body Returns #t if @1 is an absolute-URI as indicated by a
;;syntactically valid (per RFC 2396) @dfn{scheme}; otherwise returns
;;#f.

(define absolute-uri?
  (let ([re (regexp (string-append "^" SCHEME-PARSE-PATT OPAQUE-PATT "$"))])
    (lambda (str)
      (and-let* ([lst (string-match re str)])
        (and (= 3 (length lst))
             (first lst)
             (second lst)
             (third lst)
             #t) ) ) ) )

;;

(define (uri-split-path-null? path-list)
  (or (not path-list)
      (null? path-list)
      (root-split-path? path-list)) )

;;

(define (uri-path-null? uri)
  (uri-split-path-null? (uri-path uri)) )

;; Reader & Printer

(define-record-printer (uri obj port)
  (fprintf port "#,(uri ~S)" (uri->string obj)) )

(define-reader-ctor 'uri
  (lambda (str)
    (uri str) ) )

;;;
;;; Scheme Specific Setup
;;;

(set! *default-scheme-specific-class*
  (make-scheme-specific 'default-scheme
    #:fielder
    (lambda (uri key)
      (define (%authority uri)
        (first (uri-specific uri)) )
      (define (%userinfo uri)
        (or (uri$scheme-data uri)
            (let ([ui EMPTY-USERINFO-LIST])
              (and-let* ([auth (%authority uri)]
                         [str-ui (first auth)])
                (set! ui (split-userinfo str-ui)) )
              (uri$scheme-data-set! uri ui)
              ui ) ) )
      (switch key
        ['authority
          (%authority uri)]
        ['path
          (second (uri-specific uri))]
        ['query
          (third (uri-specific uri))]
        ['userinfo
          (let ([ui (%userinfo uri)])
            (values (first ui) (second ui)) )]
        ['username
          (first (%userinfo uri))]
        ['password
          (second (%userinfo uri))]
        ['host
          (and-let* ([auth (%authority uri)])
            (second auth) )]
        ['port
          (and-let* ([auth (%authority uri)])
            (third auth) )]
        [else
          (warning 'default-scheme "unknown field" key)
          (void) ] ) )
    #:parser
    (lambda (#!key scheme specific fragment authority path query)
      ;
      ; Scheme is already "valid", assuming no monkey business.
      ;
      ; Specific (overrides authority, path, and query)
      (when specific
        (unless (and (list? specific)
                     (= 3 (length specific)))
          (error 'make-uri "invalid specific" specific) )
        (set! authority (first specific))
        (set! path (second specific))
        (set! query (third specific)) )
      ;
      ; Authority
      (when (string? authority)
        (set! authority (split-authority authority 'make-uri)))
      (when (null? authority)
        (set! authority #f) )
      (unless (or (not authority)
                  (and (pair? authority)
                       (= 3 (length authority))
                       (or (not (first authority))
                           (string? (first authority)))
                       (or (not (second authority))
                           (string? (second authority)))
                       (or (not (third authority))
                           (integer? (third authority)))))
        (error 'make-uri "invalid authority" authority))
      ;
      ; Path
      (when (string? path)
        (set! path (uri-split-path path)))
      (when (null? path)
        (set! path #f))
      (unless (or (not path)
                  (and (list? path)
                       (every string? path)))
        (error 'make-uri "invalid path" path))
      ;
      ; Query
      (when (string? query)
        (set! query (split-query query (uri-missing-query-value))))
      (when (null? query)
        (set! query #f))
      (unless (or (not query)
                  (and (list? query)
                       (every query-pair? query)))
        (error 'make-uri "invalid query" query))
      ;
      ; Fragment
      (unless (or (not fragment)
                  (string? fragment))
        (error 'make-uri "invalid fragment" fragment))
      ;
      ; Ok, return the parsed argument list
      (list scheme (list authority path query) fragment) )
    #:encoder
    (lambda (uri)
      (let ([authority (first (uri-specific uri))]
            [path (second (uri-specific uri))]
            [query (third (uri-specific uri))])
        (string-append
          ; Scheme
          (if (uri-scheme uri)
              (string-append (symbol->string (uri-scheme uri)) ":")
              "")
          ; Authority
          (cond [(string? authority)
                  (string-append "//"
                                 (uri-encode authority RFC3986-SUB-DELIMS ":@[]"))]
                [(null? authority)
                  ""]
                [(pair? authority)
                  (uri-encode-authority authority)]
                [else
                  ""])
          ; Path
          (cond [(string? path)
                  (uri-encode path rfc3986-path-unreserved-char? #\/)]
                [(null? path)
                  ""]
                [(pair? path)
                  (uri-encode-split-path path)]
                [else
                  ""])
          ; Query
          (cond [(string? query)
                  (string-append "?" (uri-encode query rfc3986-path-unreserved-char? "/?"))]
                [(null? query)
                  ""]
                [(pair? query)
                  (string-append "?" (uri-encode-query query))]
                [else
                  ""])
          ; Fragment
          (if (uri-fragment uri)
              (string-append "#" (uri-encode (uri-fragment uri) rfc3986-path-unreserved-char? "/?"))
              "")) ) )
    #:decoder
    (lambda (str basis)
      (let ([empty-basis '(#f #f #f #f #f)]
            [split-uri
              (lambda ()
                (let ([nsdx (string-index str #\#)])
                  (let (
                      [fragment (and nsdx (substring str (add1 nsdx)))]
                      [wo-frag
                        (if nsdx
                            (and (not (zero? nsdx))
                                 (substring/shared str 0 nsdx))
                            str)])
                    (if wo-frag
                        (let* (
                            [len (string-length wo-frag)]
                            [qmdx (string-index wo-frag #\?)]
                            [query (and qmdx (substring wo-frag (add1 qmdx) len))]
                            [front (if qmdx
                                       (and (not (zero? qmdx))
                                            (substring wo-frag 0 qmdx))
                                       wo-frag)])
                          (if front
                              (let* (
                                  [cdx (string-index front #\:)]
                                  [scheme (and cdx (substring front 0 cdx))]
                                  [auth-path
                                    (let ([aidx (if cdx (add1 cdx) 0)])
                                      (substring front
                                                 (if (string-prefix? "//" front 0 2 aidx)
                                                     (+ aidx 2)
                                                     aidx)) ) ]
                                  [sdx (string-index auth-path #\/)]
                                  [aplen (string-length auth-path)])
                                (list
                                  scheme
                                  (if (uri-rfc3986-strict)
                                      (and sdx
                                           (not (zero? sdx))
                                           (substring auth-path 0 sdx) )
                                      (or (and sdx
                                               (not (zero? sdx))
                                               (substring auth-path 0 sdx) )
                                          (and (not (zero? aplen))
                                               auth-path ) ) )
                                  (if (uri-rfc3986-strict)
                                      (or (and sdx
                                               (not (= sdx aplen))
                                               (substring auth-path sdx) )
                                          (and (not (zero? aplen))
                                               auth-path ) )
                                      (and sdx
                                           (not (= sdx aplen))
                                           (substring auth-path sdx) ) )
                                  query
                                  fragment) )
                              (list #f #f #f query fragment) ) )
                        (list #f #f #f #f fragment) ) ) ) )])
        (let* ([uri-lst (split-uri)]
               [bas-lst
                (cond [(or (not basis)
                           (null? basis))
                        empty-basis]
                      [(pair? basis)
                        (let ([len (length basis)])
                          (cond [(= 3 len)
                                  `(,(first basis) ,@(second basis) ,(third basis))]
                                ; backwards compatibility
                                [(= 5 len)
                                  basis]
                                [else
                                  (error 'uri->list "invalid basis" basis)]) )]
                      [else
                        (error 'uri "invalid basis" basis)])]
               [uri-args
                (apply
                  (lambda (b-scheme b-authority b-path b-query b-fragment)
                    (apply
                      (lambda (scheme authority path query fragment)
                        (let ([uri-empty?
                                (and (not scheme)
                                     (not authority)
                                     (not path)
                                     (not query))])
                          (list
                            ; Scheme
                            (if scheme
                                (if (uri-scheme-symbol scheme)
                                    (string-ci->symbol scheme)
                                    (begin
                                      (warning 'uri->list "invalid scheme" scheme)
                                      b-scheme))
                                b-scheme)
                            ; Specific
                            (list
                              ; Authority
                              (if authority
                                  (uri-decode-authority authority)
                                  b-authority)
                              ; Path
                              (if uri-empty?
                                  (or b-path
                                      ROOT-SPLIT-PATH)
                                  (and path
                                       (uri-decode-path path
                                                        (and (not authority)
                                                             (not scheme)
                                                             b-path)) ) )
                              ; Query
                              (or (and query
                                       (uri-decode-query query))
                                  (and uri-empty?
                                       b-query)))
                            ; Fragment
                            (or (and fragment
                                     (uri-decode fragment))
                                (and uri-empty?
                                     b-fragment))) ) )
                      uri-lst) )
                   bas-lst)])
          (and (not (equal? uri-args EMPTY-URI-ARGS))
               uri-args) ) ) ) ) )

;;;
;;; SLIB URI Compatibility (Deprecated)
;;;

(define uri-scheme-specific uri-specific)

;;@args uri-str base-tree
;;@args uri-str
;;
;;Returns a list of 5 elements corresponding to the parts
;;(@var{scheme} @var{authority} @var{path} @var{query} @var{fragment})
;;of string @1.  Elements corresponding to absent parts are #f.
;;
;;The @var{path} is a list of strings.  If the first string is empty,
;;then the path is absolute; otherwise relative.  The optional @2 is a
;;tree as returned by @0; and is used as the base address for relative
;;URIs.
;;
;;If the @var{authority} component is a
;;@dfn{Server-based Naming Authority}, then it is a list of the
;;@var{userinfo}, @var{host}, and @var{port} strings (or #f).  For other
;;types of @var{authority} components the @var{authority} will be a
;;string.
;;
;;@example
;;(uri->tree "http://www.ics.uci.edu:80/pub/ietf/uri/#Related")
;;@result{}
;;(http (#f "www.ics.uci.edu" 80) ("" "pub" "ietf" "uri" "") #f #f "Related")
;;@end example

(define (uri->tree uri-str #!optional base-tree)
  (let ([lst (uri->list uri-str
                        (and base-tree
                             `(,(first base-tree)
                               (,(second base-tree) ,(third base-tree) ,(fourth base-tree))
                               ,(fifth base-tree))))])
    `(,(first lst) ,@(second lst) ,(third lst)) ) )

;;@args scheme authority path query fragment
;;Returns a Uniform Resource Identifier string from unencoded
;;component arguments.

(define (make-encoded-uri scheme authority path query fragment)
  ((scheme-specific-class scheme) 'encode
    (*make-uri scheme (list authority path query) fragment)) )

(define uri-make-path uri-encode-split-path)

(define path->uri pathname->uri)

;;@args str chr
;;Returns a list of @1 split at each occurrence of @2.  @2 does not
;;appear in the returned list of strings.

(define (uri-split-fields str chr)
  ;FIXME What is a CR doing in a URI!
  #;(string-split str (string chr) #t)
  (let loop ([str str] [lst '()])
    (let ([idx (string-index str chr)])
      (if idx
          (loop
            (substring/shared str (add1 idx))
            (cons
              (substring str 0
                (if (and (positive? idx)
                         (char=? #\return (string-ref str (sub1 idx))))
                    (sub1 idx)
                    idx))
              lst))
          (reverse! (cons str lst)) ) ) ) )

;;@body @1 is a path-list as returned by @code{uri-split-fields}.  @0
;;returns a list of items uri-decoded and coerced to types @2.

(define (uri-path->keys path-list ptypes)
  (and (not (uri-split-path-null? path-list))
       (let ([path (uri-decode-split-path path-list)])
         (and (= (length path) (length ptypes))
              (map coerce path ptypes)))) )

;;@body Returns #t if @1 is a fully specified pathname (does not
;;depend on the current working directory); otherwise returns #f.

(define (absolute-path? file-name)
  (and (string? file-name)
       (positive? (string-length file-name))
       (memv (string-ref file-name 0) '(#\\ #\/))) )

;;@body Returns #t if changing directory to @1 would leave the current
;;directory unchanged; otherwise returns #f.

(define (null-directory? str)
  (member str '("" "." "./" ".\\")) )

;;@body Returns #t if the string @1 contains characters used for
;;specifying glob patterns, namely @samp{*}, @samp{?}, or @samp{[}.

(define (glob-pattern? str)
  (let loop ([idx (sub1 (string-length str))])
    (and (not (negative? idx))
         (case (string-ref str idx)
           [(#\* #\[ #\?) #t]
           [else (loop (sub1 idx))])) ) )

